@@ -35,42 +35,12 @@ import socket
 import urllib.parse
 from collections.abc import Iterable, Iterator
 from email.utils import parseaddr
-from typing import Optional
+from dulwich.web import make_wsgi_chain
+from dulwich.server import DictBackend
+from itertools import takewhile
 
 import jinja2
 
-try:
-    import systemd.daemon
-except ImportError:
-    systemd_imported = False
-
-    def get_systemd_listen_sockets() -> list[socket.socket]:
-        raise NotImplementedError
-else:
-    systemd_imported = True
-
-    def get_systemd_listen_sockets() -> list[socket.socket]:
-        socks = []
-        for fd in systemd.daemon.listen_fds():
-            for family in (
-                socket.AF_UNIX,  # type: ignore
-                socket.AF_INET,
-                socket.AF_INET6,
-            ):
-                if systemd.daemon.is_socket(
-                    fd, family=family, type=socket.SOCK_STREAM, listening=True
-                ):
-                    sock = socket.fromfd(fd, family, socket.SOCK_STREAM)
-                    socks.append(sock)
-                    break
-            else:
-                raise RuntimeError(
-                    "socket family must be AF_INET, AF_INET6, or AF_UNIX; "
-                    "socket type must be SOCK_STREAM; and it must be listening"
-                )
-        return socks
-
-
 from xandikos import __version__ as xandikos_version
 from xandikos import (
     access,
@@ -108,6 +78,40 @@ from .icalendar import CalendarFilter, ICalendarFile
 from .store.git import GitStore, TreeGitStore
 from .vcard import VCardFile
 
+logger = logging.getLogger(__name__)
+
+try:
+    import systemd.daemon
+except ImportError:
+    systemd_imported = False
+
+    def get_systemd_listen_sockets() -> list[socket.socket]:
+        raise NotImplementedError
+else:
+    systemd_imported = True
+
+    def get_systemd_listen_sockets() -> list[socket.socket]:
+        socks = []
+        for fd in systemd.daemon.listen_fds():
+            for family in (
+                socket.AF_UNIX,  # type: ignore
+                socket.AF_INET,
+                socket.AF_INET6,
+            ):
+                if systemd.daemon.is_socket(
+                    fd, family=family, type=socket.SOCK_STREAM, listening=True
+                ):
+                    sock = socket.fromfd(fd, family, socket.SOCK_STREAM)
+                    socks.append(sock)
+                    break
+            else:
+                raise RuntimeError(
+                    "socket family must be AF_INET, AF_INET6, or AF_UNIX; "
+                    "socket type must be SOCK_STREAM; and it must be listening"
+                )
+        return socks
+
+
 try:
     from asyncio import to_thread  # type: ignore
 except ImportError:  # python < 3.8
@@ -130,6 +134,7 @@ STORE_CACHE_SIZE = 128
 # TODO(jelmer): Make these configurable/dynamic
 CALENDAR_HOME_SET = ["calendars"]
 ADDRESSBOOK_HOME_SET = ["contacts"]
+GIT_PATH = ".git"
 
 TEMPLATES_DIR = os.path.join(os.path.dirname(__file__), "templates")
 jinja_env = jinja2.Environment(
@@ -139,7 +144,7 @@ jinja_env = jinja2.Environment(
 
 async def render_jinja_page(
     name: str, accepted_content_languages: list[str], **kwargs
-) -> tuple[Iterable[bytes], int, Optional[str], str, list[str]]:
+) -> tuple[Iterable[bytes], int, str | None, str, list[str]]:
     """Render a HTML page from jinja template.
 
     Args:
@@ -173,7 +178,7 @@ def create_strong_etag(etag: str) -> str:
     return '"' + etag + '"'
 
 
-def extract_strong_etag(etag: Optional[str]) -> Optional[str]:
+def extract_strong_etag(etag: str | None) -> str | None:
     """Extract a strong etag from a string."""
     if etag is None:
         return etag
@@ -189,7 +194,7 @@ class ObjectResource(webdav.Resource):
         name: str,
         content_type: str,
         etag: str,
-        file: Optional[File] = None,
+        file: File | None = None,
     ) -> None:
         self.store = store
         self.name = name
@@ -198,13 +203,7 @@ class ObjectResource(webdav.Resource):
         self._file = file
 
     def __repr__(self) -> str:
-        return "{}({!r}, {!r}, {!r}, {!r})".format(
-            type(self).__name__,
-            self.store,
-            self.name,
-            self.etag,
-            self.get_content_type(),
-        )
+        return f"{type(self).__name__}({self.store!r}, {self.name!r}, {self.etag!r}, {self.get_content_type()!r})"
 
     async def get_file(self) -> File:
         if self._file is None:
@@ -343,7 +342,7 @@ class StoreBasedCollection:
         name: str,
         content_type: str,
         etag: str,
-        file: Optional[File] = None,
+        file: File | None = None,
     ) -> webdav.Resource:
         return ObjectResource(self.store, name, content_type, etag, file=file)
 
@@ -393,15 +392,36 @@ class StoreBasedCollection:
         try:
             self.store.delete_one(name, etag=extract_strong_etag(etag))
         except NoSuchItem:
-            # TODO: Properly allow removing subcollections
-            # self.get_subcollection(name).destroy()
-            shutil.rmtree(os.path.join(self.store.path, name))
+            try:
+                _subcoll = self._get_subcollection(name)
+            except KeyError:
+                # Item doesn't exist at all, raise KeyError to return 404
+                raise KeyError(name)
+            else:
+                # TODO: Properly allow removing subcollections
+                # _subcoll.destroy()
+                shutil.rmtree(os.path.join(self.store.path, name))
 
     async def create_member(
-        self, name: str, contents: Iterable[bytes], content_type: str
+        self,
+        name: str,
+        contents: Iterable[bytes],
+        content_type: str,
+        requester: str | None = None,
     ) -> tuple[str, str]:
+        # Check if member already exists and raise FileExistsError if it does
         try:
-            (name, etag) = self.store.import_one(name, content_type, contents)
+            existing_member = self.get_member(name)
+            if existing_member is not None:
+                raise FileExistsError(f"Member '{name}' already exists")
+        except KeyError:
+            # Member doesn't exist, which is what we want for create_member
+            pass
+
+        try:
+            (name, etag) = self.store.import_one(
+                name, content_type, contents, requester=requester
+            )
         except InvalidFileContents as exc:
             # TODO(jelmer): Not every invalid file is a calendar file..
             raise webdav.PreconditionFailure(
@@ -420,9 +440,9 @@ class StoreBasedCollection:
 
     def iter_differences_since(
         self, old_token: str, new_token: str
-    ) -> Iterator[tuple[str, Optional[webdav.Resource], Optional[webdav.Resource]]]:
-        old_resource: Optional[webdav.Resource]
-        new_resource: Optional[webdav.Resource]
+    ) -> Iterator[tuple[str, webdav.Resource | None, webdav.Resource | None]]:
+        old_resource: webdav.Resource | None
+        new_resource: webdav.Resource | None
         try:
             for (
                 name,
@@ -553,7 +573,7 @@ class SubscriptionCollection(StoreBasedCollection, caldav.Subscription):
         self.store.set_color(color)
 
     def get_supported_calendar_components(self):
-        return ["VEVENT", "VTODO", "VJOURNAL", "VFREEBUSY"]
+        return ["VEVENT", "VTODO", "VJOURNAL", "VFREEBUSY", "VAVAILABILITY"]
 
 
 class CalendarCollection(StoreBasedCollection, caldav.Calendar):
@@ -587,8 +607,82 @@ class CalendarCollection(StoreBasedCollection, caldav.Calendar):
     def set_calendar_timezone(self, content):
         raise NotImplementedError(self.set_calendar_timezone)
 
+    def _ensure_metadata_directory(self):
+        """Ensure .xandikos/ metadata directory exists, migrating from old .xandikos config file if needed."""
+        # Check if we already have the new directory structure by checking for config file
+        try:
+            self.store.get_file(".xandikos/config", "text/plain")
+            return  # Already migrated
+        except KeyError:
+            pass  # Need to migrate or create
+
+        # Check if we have the old .xandikos file that needs migration
+        old_config_content = None
+        try:
+            old_config_file = self.store.get_file(".xandikos", "text/plain")
+        except KeyError:
+            pass  # No old config file to migrate
+        else:
+            old_config_content = b"".join(old_config_file.content)
+            # Remove the old file
+            self.store.delete_one(".xandikos")
+
+        # Create .xandikos/ metadata directory by creating config file within it
+        if old_config_content:
+            # Migrate old config file content
+            content = [old_config_content]
+            message = "Migrate .xandikos config to metadata directory structure"
+        else:
+            # Create empty config file to establish the metadata directory
+            content = [b""]
+            message = "Create .xandikos metadata directory structure"
+
+        self.store.import_one(
+            ".xandikos/config", "text/plain", content, message=message
+        )
+
+    def get_calendar_availability(self):
+        """Get calendar availability from .xandikos/availability.ics file."""
+        try:
+            availability_file = self.store.get_file(
+                ".xandikos/availability.ics", "text/calendar"
+            )
+        except NoSuchItem:
+            raise KeyError
+
+        return b"".join(availability_file.content).decode("utf-8")
+
+    def set_calendar_availability(self, content):
+        """Set calendar availability by storing in .xandikos/availability.ics file."""
+        # Ensure .xandikos/ metadata directory exists (migrates config if needed)
+        self._ensure_metadata_directory()
+
+        if content is None:
+            # Remove availability
+            try:
+                self.store.delete_one(".xandikos/availability.ics")
+            except NoSuchItem:
+                pass  # Already removed
+        else:
+            # Validate that it's valid iCalendar data and normalize it
+            try:
+                from icalendar.cal import Calendar as ICalendar
+
+                cal = ICalendar.from_ical(content)
+            except (ValueError, UnicodeDecodeError, TypeError, KeyError) as e:
+                raise InvalidFileContents("text/calendar", content, e)
+
+            # Store the normalized form
+            normalized_content = cal.to_ical().decode("utf-8")
+            self.store.import_one(
+                ".xandikos/availability.ics",
+                "text/calendar",
+                [normalized_content.encode("utf-8")],
+                message="Update calendar availability",
+            )
+
     def get_supported_calendar_components(self):
-        return ["VEVENT", "VTODO", "VJOURNAL", "VFREEBUSY"]
+        return ["VEVENT", "VTODO", "VJOURNAL", "VFREEBUSY", "VAVAILABILITY"]
 
     def get_supported_calendar_data_types(self):
         return [("text/calendar", "1.0"), ("text/calendar", "2.0")]
@@ -665,6 +759,14 @@ class AddressbookCollection(StoreBasedCollection, carddav.Addressbook):
     def set_addressbook_color(self, color):
         self.store.set_color(color)
 
+    def addressbook_query(self, create_filter_fn):
+        from .vcard import CardDAVFilter
+
+        filter = create_filter_fn(CardDAVFilter)
+        for name, file, etag in self.store.iter_with_filter(filter=filter):
+            resource = self._get_resource(name, file.content_type, etag, file=file)
+            yield (name, resource)
+
     def get_addressbook_color(self):
         color = self.store.get_color()
         if not color:
@@ -793,11 +895,68 @@ class RootPage(webdav.Resource):
     def render(self, self_url, accepted_content_types, accepted_content_languages):
         content_types = webdav.pick_content_types(accepted_content_types, ["text/html"])
         assert content_types == ["text/html"]
+
+        # Generate CalDAV/CardDAV URLs
+        from urllib.parse import urlparse, urlunparse
+
+        parsed = urlparse(self_url)
+
+        # Determine if we're using HTTPS
+        is_secure = parsed.scheme == "https"
+
+        # Create URLs with different schemes, preserving the full path
+        caldav_url = urlunparse(
+            (
+                "caldavs" if is_secure else "caldav",
+                parsed.netloc,
+                parsed.path,
+                "",
+                "",
+                "",
+            )
+        )
+        carddav_url = urlunparse(
+            (
+                "carddavs" if is_secure else "carddav",
+                parsed.netloc,
+                parsed.path,
+                "",
+                "",
+                "",
+            )
+        )
+
+        # Generate DAV×5 URL for QR code
+        davx5_url = urlunparse(("davx5", parsed.netloc, parsed.path, "", "", ""))
+
+        # Try to generate QR code if qrcode is available
+        qr_code_data = None
+        try:
+            import qrcode
+        except ImportError:
+            logger.warning("qrcode package not installed; QR code generation disabled")
+        else:
+            import io
+            import base64
+
+            qr = qrcode.QRCode(version=1, box_size=10, border=4)
+            qr.add_data(davx5_url)
+            qr.make(fit=True)
+
+            img = qr.make_image(fill_color="black", back_color="white")
+            buffer = io.BytesIO()
+            img.save(buffer, format="PNG")
+            qr_code_data = base64.b64encode(buffer.getvalue()).decode()
+
         return render_jinja_page(
             "root.html",
             accepted_content_languages,
             principals=self.backend.find_principals(),
             self_url=self_url,
+            caldav_url=caldav_url,
+            carddav_url=carddav_url,
+            davx5_url=davx5_url,
+            qr_code_data=qr_code_data,
         )
 
     async def get_body(self):
@@ -979,7 +1138,7 @@ def open_store_from_path(path: str, **kwargs):
 
 class XandikosBackend(webdav.Backend):
     def __init__(
-        self, path, *, paranoid: bool = False, index_threshold: Optional[int] = None
+        self, path, *, paranoid: bool = False, index_threshold: int | None = None
     ) -> None:
         self.path = path
         self._user_principals: set[str] = set()
@@ -1049,6 +1208,70 @@ class XandikosBackend(webdav.Backend):
             except KeyError:
                 return None
 
+    async def copy_collection(
+        self, source_path: str, dest_path: str, overwrite: bool = True
+    ) -> bool:
+        """Copy a collection recursively."""
+        import shutil
+
+        source_collection = self.get_resource(source_path)
+        if source_collection is None:
+            raise KeyError(source_path)
+
+        if webdav.COLLECTION_RESOURCE_TYPE not in source_collection.resource_types:
+            raise ValueError(f"Source '{source_path}' is not a collection")
+
+        source_file_path = self._map_to_file_path(source_path)
+        dest_file_path = self._map_to_file_path(dest_path)
+
+        # Check if destination exists
+        did_overwrite = False
+        if os.path.exists(dest_file_path):
+            if not overwrite:
+                raise FileExistsError(f"Collection '{dest_path}' already exists")
+            # Remove existing destination (file or directory)
+            did_overwrite = True
+            if os.path.isdir(dest_file_path):
+                shutil.rmtree(dest_file_path)
+            else:
+                os.remove(dest_file_path)
+
+        # Copy the entire directory tree
+        shutil.copytree(source_file_path, dest_file_path)
+        return did_overwrite
+
+    async def move_collection(
+        self, source_path: str, dest_path: str, overwrite: bool = True
+    ) -> bool:
+        """Move a collection recursively."""
+        import shutil
+
+        source_collection = self.get_resource(source_path)
+        if source_collection is None:
+            raise KeyError(source_path)
+
+        if webdav.COLLECTION_RESOURCE_TYPE not in source_collection.resource_types:
+            raise ValueError(f"Source '{source_path}' is not a collection")
+
+        source_file_path = self._map_to_file_path(source_path)
+        dest_file_path = self._map_to_file_path(dest_path)
+
+        # Check if destination exists
+        did_overwrite = False
+        if os.path.exists(dest_file_path):
+            if not overwrite:
+                raise FileExistsError(f"Collection '{dest_path}' already exists")
+            did_overwrite = True
+            # Remove existing destination (file or directory)
+            if os.path.isdir(dest_file_path):
+                shutil.rmtree(dest_file_path)
+            else:
+                os.remove(dest_file_path)
+
+        # Move the entire directory tree
+        shutil.move(source_file_path, dest_file_path)
+        return did_overwrite
+
 
 class XandikosApp(webdav.WebDAVApp):
     """A wsgi App that provides a Xandikos web server."""
@@ -1090,6 +1313,7 @@ class XandikosApp(webdav.WebDAVApp):
                 sync.SyncTokenProperty(),
                 caldav.SupportedCalendarDataProperty(),
                 caldav.CalendarTimezoneProperty(),
+                caldav.CalendarAvailabilityProperty(),
                 caldav.MinDateTimeProperty(),
                 caldav.MaxDateTimeProperty(),
                 caldav.MaxResourceSizeProperty(),
@@ -1148,6 +1372,29 @@ class XandikosApp(webdav.WebDAVApp):
             ]
         )
 
+    async def _handle_request(self, request, environ, start_response=None):
+        if start_response and GIT_PATH in request.path.split(posixpath.sep):
+            return self._handle_git_request(
+                request,
+                environ["ORIGINAL_ENVIRON"],
+                takewhile(lambda x: x != GIT_PATH, request.path.split(posixpath.sep)),
+                start_response,
+            )
+        else:
+            return await super()._handle_request(request, environ)
+
+    def _handle_git_request(self, request, environ, path, start_response):
+        resource_path = posixpath.join("/", *path)
+        resource = self.backend.get_resource(resource_path)
+        if not isinstance(resource, StoreBasedCollection) or not isinstance(
+            resource.store, GitStore
+        ):
+            return webdav._send_not_found(request)
+
+        prefix = posixpath.join(resource_path, GIT_PATH)
+        chain = make_wsgi_chain(DictBackend({prefix: resource.store.repo}), dumb=True)
+        return chain(environ, start_response)
+
 
 def create_principal_defaults(backend, principal):
     """Create default calendar and addressbook for a principal.
@@ -1243,9 +1490,9 @@ def run_simple_server(
     defaults: bool = False,
     strict: bool = True,
     route_prefix: str = "/",
-    listen_address: Optional[str] = "::",
-    port: Optional[int] = 8080,
-    socket_path: Optional[str] = None,
+    listen_address: str | None = "::",
+    port: int | None = 8080,
+    socket_path: str | None = None,
 ) -> None:
     """Simple function to run a Xandikos server.
 
@@ -1279,8 +1526,7 @@ def run_simple_server(
         )
     if not backend.get_resource(current_user_principal):
         logging.warning(
-            "default user principal %s does not exist. "
-            "Run xandikos with --autocreate?",
+            "default user principal %s does not exist. Run xandikos with --autocreate?",
             current_user_principal,
         )
 
@@ -1319,19 +1565,8 @@ def run_simple_server(
     web.run_app(app, port=port, host=listen_address, path=socket_path)
 
 
-async def main(argv=None):  # noqa: C901
+def add_parser(parser):
     import argparse
-    import sys
-
-    from xandikos import __version__
-
-    parser = argparse.ArgumentParser(usage="%(prog)s -d ROOT-DIR [OPTIONS]")
-
-    parser.add_argument(
-        "--version",
-        action="version",
-        version="%(prog)s " + ".".join(map(str, __version__)),
-    )
 
     access_group = parser.add_argument_group(title="Access Options")
     access_group.add_argument(
@@ -1347,8 +1582,7 @@ async def main(argv=None):  # noqa: C901
         dest="listen_address",
         default="localhost",
         help=(
-            "Bind to this address. "
-            "Pass in path for unix domain socket. [%(default)s]"
+            "Bind to this address. Pass in path for unix domain socket. [%(default)s]"
         ),
     )
     access_group.add_argument(
@@ -1362,7 +1596,7 @@ async def main(argv=None):  # noqa: C901
     access_group.add_argument(
         "--metrics-port",
         dest="metrics_port",
-        default=8081,
+        default=None,
         help="Port to listen on for metrics. [%(default)s]",
     )
     access_group.add_argument(
@@ -1379,6 +1613,7 @@ async def main(argv=None):  # noqa: C901
         "--directory",
         dest="directory",
         default=None,
+        required=True,
         help="Directory to serve from.",
     )
     parser.add_argument(
@@ -1396,7 +1631,7 @@ async def main(argv=None):  # noqa: C901
         "--defaults",
         action="store_true",
         dest="defaults",
-        help=("Create initial calendar and address book. " "Implies --autocreate."),
+        help=("Create initial calendar and address book. Implies --autocreate."),
     )
     parser.add_argument(
         "--dump-dav-xml",
@@ -1411,7 +1646,7 @@ async def main(argv=None):  # noqa: C901
         "--no-strict",
         action="store_false",
         dest="strict",
-        help=("Enable workarounds for buggy CalDAV/CardDAV client " "implementations."),
+        help=("Enable workarounds for buggy CalDAV/CardDAV client implementations."),
         default=True,
     )
     parser.add_argument("--debug", action="store_true", help="Print debug messages")
@@ -1419,12 +1654,9 @@ async def main(argv=None):  # noqa: C901
     # and are generally just meant for developers.
     parser.add_argument("--paranoid", action="store_true", help=argparse.SUPPRESS)
     parser.add_argument("--index-threshold", type=int, help=argparse.SUPPRESS)
-    options = parser.parse_args(argv)
 
-    if options.directory is None:
-        parser.print_usage()
-        sys.exit(1)
 
+async def main(options, parser):
     if options.dump_dav_xml:
         # TODO(jelmer): Find a way to propagate this without abusing
         # os.environ.
@@ -1461,11 +1693,12 @@ async def main(argv=None):  # noqa: C901
         )
     if not backend.get_resource(options.current_user_principal):
         logging.warning(
-            "default user principal %s does not exist. "
-            "Run xandikos with --autocreate?",
+            "default user principal %s does not exist. Run xandikos with --autocreate?",
             options.current_user_principal,
         )
 
+    logging.info("Xandikos %s", ".".join(map(str, xandikos_version)))
+
     main_app = XandikosApp(
         backend,
         current_user_principal=options.current_user_principal,
@@ -1503,13 +1736,13 @@ async def main(argv=None):  # noqa: C901
         parser.error("Metrics port cannot be the same as the main port")
 
     app = web.Application()
-    if options.metrics_port:
+    if options.metrics_port is not None:
         metrics_app = web.Application()
         try:
             from aiohttp_openmetrics import metrics, metrics_middleware
         except ModuleNotFoundError:
             logging.warning(
-                "aiohttp-openmetrics not found; " "/metrics will not be available."
+                "aiohttp-openmetrics not found; /metrics will not be available."
             )
         else:
             app.middlewares.insert(0, metrics_middleware)
@@ -1543,7 +1776,7 @@ async def main(argv=None):  # noqa: C901
             import dbus  # noqa: F401
         except ImportError:
             logging.error(
-                "Please install python-avahi and python-dbus for " "avahi support."
+                "Please install python-avahi and python-dbus for avahi support."
             )
         else:
             avahi_register(options.port, options.route_prefix)
@@ -1554,7 +1787,7 @@ async def main(argv=None):  # noqa: C901
     if metrics_app:
         metrics_runner = web.AppRunner(metrics_app)
         await metrics_runner.setup()
-        # TODO(jelmer): Allow different metrics listen addres?
+        # TODO(jelmer): Allow different metrics listen address?
         sites.append(web.TCPSite(metrics_runner, listen_address, options.metrics_port))
     # Use systemd sockets first and only if not present use the socket path or
     # address from --listen-address.
@@ -1567,18 +1800,47 @@ async def main(argv=None):  # noqa: C901
 
     import signal
 
-    # Set SIGINT to default handler; this appears to be necessary
-    # when running under coverage.
-    signal.signal(signal.SIGINT, signal.SIG_DFL)
+    # Set up graceful shutdown handling
+    loop = asyncio.get_running_loop()
+    shutdown_event = asyncio.Event()
+
+    def signal_handler(signum, frame):
+        logging.info("Received signal %s, shutting down gracefully...", signum)
+        # Use call_soon_threadsafe to safely set the event from signal handler
+        loop.call_soon_threadsafe(shutdown_event.set)
+
+    # Register signal handlers for graceful shutdown
+    signal.signal(signal.SIGTERM, signal_handler)
+    signal.signal(signal.SIGINT, signal_handler)
 
     for site in sites:
         await site.start()
 
-    while True:
-        await asyncio.sleep(3600)
+    # Wait for shutdown signal
+    try:
+        await shutdown_event.wait()
+    except KeyboardInterrupt:
+        logging.info("Received KeyboardInterrupt, shutting down gracefully...")
+
+    # Cleanup: stop all sites and runners
+    logging.info("Stopping web servers...")
+    for site in sites:
+        await site.stop()
+
+    await runner.cleanup()
+    if metrics_app:
+        await metrics_runner.cleanup()
+
+    logging.info("Shutdown complete.")
 
 
 if __name__ == "__main__":
     import sys
 
-    sys.exit(asyncio.run(main(sys.argv[1:])))
+    import argparse
+
+    parser = argparse.ArgumentParser(usage="%(prog)s [options]")
+    add_parser(parser)
+    args = parser.parse_args(sys.argv[1:])
+
+    sys.exit(asyncio.run(main(args, parser)))
